type boisson = 
	| Eau
	| Jus of string
	| BreizhCola of bool (* true si light, false sinon *)
	| Cocktail of boisson * boisson * float (* deux boissons b1 b2, et la proportion de b1 *)

(* prix au litre de b*)
let rec prix (b: boisson) : float = match b with
	| Eau -> 0.
	| BreizhCola _ -> 1.
	| Jus "ramboutan" -> 5.3
	| Jus _ -> 3.
	| Cocktail (b1, b2, x) -> x *. prix b1 +. (1. -. x) *. prix b2

(* boisson de la forme Cocktail(0.5, b1, Cocktail(0.5, b2, ....))
   où l_boissons = [b1; b2; ... bn] *)
let rec shaker (l_boissons: boisson list) : boisson =
	match l_boissons with
	| [] -> failwith "shaker vide"
	| x :: [] -> x 
	| x :: q -> Cocktail (x, shaker q, 0.5)

(* chaîne représentant b une boisson de base *)
let string_of_boisson (b: boisson) : string = 
	match b with 
	| Cocktail _ -> failwith "Pas une boisson de base"
	| Eau -> "Eau"
	| Jus fruit -> 
		(* Dire "jus d'orange" et pas "jus de orange" *)
		begin match fruit.[0] with 
		| 'a' | 'o' | 'e' | 'i' | 'u' | 'h' -> "Jus d'" ^ fruit
		| _ -> "Jus de " ^ fruit
		end   
	| BreizhCola true -> "Breizh Cola Light"
	| BreizhCola false -> "Breizh Cola"


(* Liste des ingrédients nécessaires pour
	préparer v litres de b, avec chaque proportion multipliée par p *)
let rec ingredients (b: boisson) (v: float) : (boisson*float) list =
	match b with 
		| Eau | Jus _ | BreizhCola _ -> [(b, v)]
		| Cocktail (b1, b2, x) ->
			(* l'opérateur @ fait une concaténation de listes.  *) 
			ingredients b1 (v*.x) @ 
			ingredients b2 (v*.(1. -. x))
(*
Attention, en terme de complexité, l'opération l1 @ l2 a 
un coût O(n1) avec n1 la taille de l1.
Ainsi, si l'on considère une boisson de la forme
Cocktail(Cocktail(....Cocktail(Cocktail(Eau, Eau), Eau), ... )),
avec n occurrences du constructeur Cocktail, la fonction
va effectuer O(n + (n-1) + ... + 1) opérations juste avec
la concaténation, soit O(n²). Voici une version linéaire
de la fonction: elle prend en entrée un paramètre supplémentaire,
une liste servant à stocker les ingrédients déjà croisé.
*)

(* Renvoie les ingrédients de v litres de b, ainsi que la liste l *)
let rec ingredients2 (b: boisson) (v: float) (l: (boisson*float) list): (boisson*float) list =
	match b with 
		| Eau | Jus _ | BreizhCola _ -> (b, v) :: l
		| Cocktail (b1, b2, x) ->
			(* ajouter les ingrédients de b1 à l -> obtenir l1 *)
			let l1 = ingredients2 b1 (v*.x) l in 
			(* ajouter les ingrédients de b2 à l -> obtenir l2 *)
			let l2 = ingredients2 b2 (v*.(1. -. x)) l1 in
			l2




(* Regroupe les éléments de l par première composante,  en sommant
   sur la deuxième composante. Précondition: l est triée, et
   en particulier, les éléments de même première composante sont
   tous consécutifs *)
let agreg_sum (l: ('a * float) list) : ('a * float) list =
	(* agreg_sum_curr ll curr curr_p renvoie agreg_sum ll, en ayant déjà
	   vu une quantité curr_p de l'élément curr. *)
	let rec agreg_sum_curr 
		(ll: ('a * float) list) (curr: 'a) (curr_p: float) : ('a * float) list  = 
		match ll with 
		| [] -> [(curr, curr_p)]
		| (x, p) :: q -> if x = curr then agreg_sum_curr q curr (curr_p +. p)
						 else (curr, curr_p) :: agreg_sum_curr q x p
	in match l with 
	| [] -> []
	| (x, p) :: q -> agreg_sum_curr q x p


(** Tri fusion **)

let rec fusion (l1: 'a list) (l2: 'a list) : 'a list =
	match l1, l2 with 
	| [], _ -> l2
	| _, [] -> l1
	| x1 :: q1, x2 :: q2 -> 
		 if x1 < x2 then 
		 	x1 :: fusion q1 l2
		 else
		 	x2 :: fusion l1 q2

let rec separer (l: 'a list) : ('a list * 'a list) =
	match l with 
	| [] -> ([], [])
	| [x] -> ([x], [])
	| x :: y :: q ->
		let (l1, l2) = separer q in (x::l1, y::l2)

let rec tri_fusion (l: 'a list) : 'a list =
	match l with 
	| [] | [_] -> l
	| _ -> let l1, l2 = separer l in
			 fusion (tri_fusion l1) (tri_fusion l2) 

(* Affiche la recette de b *)
let recette (b: boisson) : unit =
	let l = ingredients b 1. in
	let l_regroupee = agreg_sum (tri_fusion l) in
	print_string "Recette pour 1L:\n";
	(* Affiche chaque couple (bb, p) de ll sous la forme:
	    "p mL de bb" + retour à la ligne
	   Chaque boisson bb de ll doit être une boisson de base. *)
	let rec aff_liste (ll: (boisson * float) list) : unit = 
		match ll with 
		| [] -> ()
		| (bb, p) :: q -> begin 
			print_int (int_of_float (p *. 1000.)); 
			print_string " mL ";
			print_string (string_of_boisson bb);
			print_newline ();
			aff_liste q
		end
	in aff_liste l_regroupee

let test () =
	let b =
	Cocktail(
		Cocktail(
			Jus "raisin",
			Cocktail(
				BreizhCola true,
				Eau,
				0.5
			),
			0.4
		),
		Cocktail(
			Jus "raisin",
			BreizhCola false,
			0.7
		), 
		0.1
	)
	in recette b

